/*-
 * Copyright (c) 2000 Munehiro Matsuda
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$FreeBSD$
 */

#include <sys/param.h>
#include <sys/ioctl.h>

#include <dev/acpica/acpiio.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define ACPIDEV	"/dev/acpi"

static void
usage(void)
{
	fprintf(stderr, "usage: acpibatt [-vd?]\n");
	exit(1);
}

static int
acpi_acad_status(void)
{
	int	 fd;
	int	 status;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	if (ioctl(fd, ACPIIO_ACAD_GET_STATUS, &status) == -1) {
		err(1, "Failed to get AC-Line info");
	}
	close(fd);

	return(status);
}

static void
acpi_batt_get_units(int *unitsp)
{
	int	fd;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	if (ioctl(fd, ACPIIO_BATT_GET_UNITS, unitsp) == -1) {
		err(1, "Failed to get number of batteries");
	}
	close(fd);
}

static void
acpi_batt_get_battdesc(int unit, struct acpi_battdesc *battdesc)
{
	int	 fd;
	union	 acpi_battery_ioctl_arg arg;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	arg.unit = unit;
	if (ioctl(fd, ACPIIO_BATT_GET_BATTDESC, &arg) == -1) {
		err(1, "Failed to get battery desc");
	}
	close(fd);

	bcopy(&arg.battdesc, battdesc, sizeof(struct acpi_battdesc));
}

static void
acpi_batt_get_battinfo(int unit, struct acpi_battinfo *battinfo)
{
	int	 fd;
	union	 acpi_battery_ioctl_arg arg;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	arg.unit = unit;
	if (ioctl(fd, ACPIIO_BATT_GET_BATTINFO, &arg) == -1) {
		err(1, "Failed to get battery info");
	}
	close(fd);

	bcopy(&arg.battinfo, battinfo, sizeof(struct acpi_battinfo));
}

static void
acpi_batt_get_cmbat_info(int unit, struct acpi_bif *bifp, struct acpi_bst *bstp)
{
	int	 fd;
	union	 acpi_battery_ioctl_arg arg;

	fd  = open(ACPIDEV, O_RDONLY);
	if (fd == -1) {
		err(1, "%s", ACPIDEV);
	}
	arg.unit = unit;
	if (ioctl(fd, ACPIIO_CMBAT_GET_BIF, &arg) == -1) {
		err(1, "Failed to get ACPI _BIF info");
	}
	bcopy(&arg.bif, bifp, sizeof(struct acpi_bif));

	arg.unit = unit;
	if (ioctl(fd, ACPIIO_CMBAT_GET_BST, &arg) == -1) {
		err(1, "Failed to get ACPI _BST info");
	}
	bcopy(&arg.bst, bstp, sizeof(struct acpi_bst));
	close(fd);
}

static void
acpi_batt_print_additional_info(int unit, int verbose, int debug)
{
	struct	 acpi_battdesc battdesc;
	struct	 acpi_bif bif;
	struct	 acpi_bst bst;
	char	*power_units[]= { "mWh", "mAh" };

	acpi_batt_get_battdesc(unit, &battdesc);

	switch (battdesc.type) {
	case ACPI_BATT_TYPE_CMBAT:
		if (verbose || debug) {
			acpi_batt_get_cmbat_info(battdesc.phys_unit, &bif, &bst);
		}
			
		if (verbose) {
			printf("\tBattery output: ");
			if (bst.volt == 0xffffffff) {
				printf("unknown\n");
			} else {
				printf("%gV\n", bst.volt / 1000.0);
			}

			printf("\tBattery model : %s\n", bif.model);
			printf("\tBattery serial: %s\n", bif.serial);
			printf("\tBattery type  : %s\n", bif.type);
			printf("\tBattery tech  : %s\n", bif.btech > 0 ?
			    "Secondary (rechargeable)" :
			    "Primary (non-rechargeable)");
			printf("\tBattery oeminfo: %s\n", bif.oeminfo);
		}

		/* Debugging informations */
		if (debug > 0) {
			char *u = power_units[bif.unit & 0x01];

			printf("\t_BIF info\n");
			printf("\t  unit : %d (%s)\n", bif.unit, u);
			printf("\t  dcap : %5d%s\n", bif.dcap, u);
			printf("\t  lfcap: %5d%s\n", bif.lfcap, u);
			printf("\t  wcap : %5d%s\n", bif.wcap, u);
			printf("\t  lcap : %5d%s\n", bif.lcap, u);
			printf("\t  gra1 : %5d%s\n", bif.gra1, u);
			printf("\t  gra2 : %5d%s\n", bif.gra2, u);
			printf("\t  dvol : %5dmV\n", bif.dvol);
#ifdef DEBUG
			printf("\t  btech: %d\n", bif.btech);
			printf("\t  model  : %s\n", bif.model);
			printf("\t  serial : %s\n", bif.serial);
			printf("\t  type   : %s\n", bif.type);
			printf("\t  oeminfo: %s\n", bif.oeminfo);
#endif
			printf("\t_BST info\n");
			printf("\t  state: %d\n", bst.state);
			printf("\t  rate : %5d%s\n", bst.rate, u);
			printf("\t  cap  : %5d%s\n", bst.cap, u);
			printf("\t  volt : %5dmV\n", bst.volt);
		}
		break;

	default:
		break;
	}
}

int
main(int argc, char *argv[])
{
	int	 i, c;
	int	 debug;
	int	 units;
	int	 verbose;
	struct	 acpi_battinfo battinfo;
	char	*status[ACPI_BATT_STAT_MAX + 1] = {
		"normal",
		"dis-charging",
		"charging",
		NULL,
		"critical",
		"critical and dis-charging",
		"critical and charging",
		"not present"
	};

	verbose = debug = 0;
	while ((c = getopt(argc, argv, "dg:s:v?")) != -1) {
		switch (c) {
		case 'd':
			debug++;
			break;
		case 'v':
			verbose++;
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	/* Print Battery related informations */
	printf("ACPI Battey info:\n");
	printf("AC Line status: %s\n",
	    acpi_acad_status() ? "on-line" : "off-line");

	acpi_batt_get_battinfo(-1, &battinfo);

	printf("Remaining battery life: ");
	if (battinfo.cap == -1) {
		printf("unknown\n");
	} else {
		printf("%d%%\n", battinfo.cap);
	}

	printf("Remaining battery time: ");
	if (battinfo.min == -1) {
		printf("unknown\n");
	} else {
		int t, h, m;

		t = battinfo.min;
		m = t % 60;
		t /= 60;
		h = t;
		printf("%2d:%02d:00\n", h, m);
	}

	acpi_batt_get_units(&units);

	printf("Number of batteries: %d\n", units);
	for (i = 0; i < units; i++) {
		acpi_batt_get_battinfo(i, &battinfo);

		printf("Battery %d:\n", i);

		printf("\tBattery status: %s\n", status[battinfo.state]);
		if (battinfo.state == ACPI_BATT_STAT_NOT_PRESENT)
			continue;

		printf("\tRemaining battery life: %d%%\n", battinfo.cap);

		printf("\tRemaining battery time: ");
		if (battinfo.min > 0) {
			int t, h, m;

			t = battinfo.min;
			m = t % 60;
			t /= 60;
			h = t;
			printf("%2d:%02d:00\n", h, m);
		} else {
			printf("unknown\n");
		}

		/* Additional informations */
		acpi_batt_print_additional_info(i, verbose, debug);
	}

	return (0);
}

